
#include <cassert>

#include <cstddef>
using std::size_t;

#include <cmath>
using std::fabs;

#include <iostream>
using std::cerr;
using std::endl;

#include <vector>
using std::vector;

#include <numeric>
using std::accumulate;

#include "mtrand.hpp"
mtrand R;

#include "moments.hpp"

// global variables
vector<int> spin;
size_t L;
double T;

#ifdef USE_LINEAR_CHAIN
size_t site_count(size_t L) { return L; }
size_t right(size_t i) { return (i+1)%L; }
size_t left(size_t i) { return (L+i-1)%L; }
int neighbour_sum(size_t i) { return spin[right(i)] + spin[left(i)]; }
#endif

#ifdef USE_SQUARE_LATTICE
size_t site_count(size_t L) { return L*L; }
size_t right(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return (col+1)%L + row*L;
}
size_t left(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return (L+col-1)%L + row*L;
}
size_t up(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return col + ((row+1)%L)*L;
}
size_t down(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return col + ((L+row-1)%L)*L;
}
int neighbour_sum(size_t i) 
{ 
	return spin[right(i)] + spin[left(i)]
	       + spin[up(i)] + spin[down(i)];
}
#endif

void initialize(size_t L_, double T_)
{
	L = L_;
	T = T_;
	spin.resize(site_count(L_));
	for (size_t i = 0; i < spin.size(); ++i)
		spin[i] = 1; // ferromagnetic initial state
		//spin[i] = ( R() < 0.5 ? 1 : -1); // random initial state
}

double w;
vector<size_t> sites_to_flip;

void attempt_single_flip(void)
{
	assert(sites_to_flip.empty());
	const size_t i = R.randIntExc(spin.size());
	sites_to_flip.push_back(i);
	w = exp(-2.0*spin[i]*neighbour_sum(i)/T);
}

void attempt_cluster_flip(void);

void update(void)
{
	if (w > 1.0 or w > R())
	{
		// accept
		while (!sites_to_flip.empty())
		{
			spin[ sites_to_flip.back() ] *= -1;
			sites_to_flip.pop_back();
		}
	}
	else
	{
		// reject
		sites_to_flip.clear();		
	}
}

void measure(moments &O)
{
	const int M = accumulate(spin.begin(),spin.end(),0);
	const double M1 = fabs(double(M));
	const double M2 = M1*M1;
	const double M4 = M2*M2;
	O.abs_first += M1;
	O.second += M2;
	O.fourth += M4;
}
